@@ -2,12 +2,51 @@ import jax
 import jax.numpy as jnp
 from jax import lax
 from gymnax.environments import environment, spaces
-
-from typing import Tuple
+from typing import Tuple, Optional
 import chex
+from flax import struct
 
-Array = chex.Array
-PRNGKey = chex.PRNGKey
+
+@struct.dataclass
+class EnvState:
+    oxygen: int
+    sub_x: int
+    sub_y: int
+    sub_or: int
+    f_bullet_count: int
+    f_bullets: chex.Array
+    e_bullet_count: int
+    e_bullets: chex.Array
+    e_fish_count: int
+    e_fish: chex.Array
+    e_subs_count: int
+    e_subs: chex.Array
+    diver_count: int
+    divers: chex.Array
+    e_spawn_speed: int
+    e_spawn_timer: int
+    d_spawn_timer: int
+    move_speed: int
+    ramp_index: int
+    shot_timer: int
+    surface: int
+    time: int
+    terminal: bool
+
+
+@struct.dataclass
+class EnvParams:
+    ramping: bool = True
+    ramp_interval: int = 100
+    init_spawn_speed: int = 20
+    init_move_interval: int = 5
+    max_oxygen: int = 200
+    diver_spawn_speed: int = 30
+    shot_cool_down: int = 5
+    enemy_shot_interval: int = 10
+    enemy_move_interval: int = 5
+    diver_move_interval: int = 5
+    max_steps_in_episode: int = 1000
 
 
 class MinSeaquest(environment.Environment):
@@ -58,38 +97,30 @@ class MinSeaquest(environment.Environment):
             self.action_set = self.full_action_set
 
     @property
-    def default_params(self):
+    def default_params(self) -> EnvParams:
         # Default environment parameters
-        return {
-            "ramp_interval": 100,
-            "max_oxygen": 200,
-            "init_spawn_speed": 20,
-            "diver_spawn_speed": 30,
-            "init_move_interval": 5,
-            "shot_cool_down": 5,
-            "enemy_shot_interval": 10,
-            "enemy_move_interval": 5,
-            "diver_move_interval": 5,
-            "ramping": 1,
-            "max_steps_in_episode": 1000,
-        }
+        return EnvParams()
 
     def step_env(
-        self, key: PRNGKey, state: dict, action: int, params: dict
-    ) -> Tuple[Array, dict, float, bool, dict]:
+        self, key: chex.PRNGKey, state: EnvState, action: int, params: EnvParams
+    ) -> Tuple[chex.Array, EnvState, float, bool, dict]:
         """Perform single timestep state transition."""
         # If timer is up spawn enemy and divers [always sample]
         key_enemy, key_diver = jax.random.split(key)
         spawned_enemy = spawn_enemy(key_enemy)
-        spawn_enemy_cond = state["e_spawn_timer"] == 0
-        state["e_spawn_timer"] = lax.select(
-            spawn_enemy_cond, state["e_spawn_speed"], state["e_spawn_timer"]
+        spawn_enemy_cond = state.e_spawn_timer == 0
+        state = state.replace(
+            e_spawn_timer=lax.select(
+                spawn_enemy_cond, state.e_spawn_speed, state.e_spawn_timer
+            )
         )
 
         spawned_diver = spawn_diver(key_diver)
-        spawn_diver_cond = state["d_spawn_timer"] == 0
-        state["d_spawn_timer"] = lax.select(
-            spawn_diver_cond, state["d_spawn_speed"], state["d_spawn_timer"]
+        spawn_diver_cond = state.d_spawn_timer == 0
+        state = state.replace(
+            d_spawn_timer=lax.select(
+                spawn_diver_cond, state.d_spawn_speed, state.d_spawn_timer
+            )
         )
 
         # Sequentially go through substate and update the state
@@ -101,153 +132,114 @@ class MinSeaquest(environment.Environment):
         state, reward = step_e_bullets(state, reward)
         state, reward = step_timers(state, reward)
         # Check game condition & no. steps for termination condition
-        state["time"] += 1
+        state.replace(time=state.time + 1)
         done = self.is_terminal(state, params)
-        state["terminal"] = done
+        state = state.replace(terminal=done)
         info = {"discount": self.discount(state, params)}
         return (
             lax.stop_gradient(self.get_obs(state, params)),
             lax.stop_gradient(state),
-            reward,
+            reward.astype(jnp.float32),
             done,
             info,
         )
 
-    def reset_env(self, key: PRNGKey, params: dict) -> Tuple[Array, dict]:
+    def reset_env(
+        self, key: chex.PRNGKey, params: EnvParams
+    ) -> Tuple[chex.Array, EnvState]:
         """Reset environment state by sampling initial position."""
-        state = {
-            "oxygen": params["max_oxygen"],
-            "sub_x": 5,
-            "sub_y": 0,
-            "sub_or": 0,
-            "f_bullet_count": 0,
-            "f_bullets": jnp.zeros((100, 3), dtype=jnp.int32),
-            "e_bullet_count": 0,
-            "e_bullets": jnp.zeros((100, 3), dtype=jnp.int32),
-            "e_fish_count": 0,
-            "e_fish": jnp.zeros((100, 5), dtype=jnp.int32),
-            "e_subs_count": 0,
-            "e_subs": jnp.zeros((100, 5), dtype=jnp.int32),
-            "diver_count": 0,
-            "divers": jnp.zeros((100, 4), dtype=jnp.int32),
-            "e_spawn_speed": params["init_spawn_speed"],
-            "e_spawn_timer": params["init_spawn_speed"],
-            "d_spawn_timer": params["diver_spawn_speed"],
-            "move_speed": params["init_move_interval"],
-            "ramp_index": 0,
-            "shot_timer": 0,
-            "surface": 1,
-            "time": 0,
-            "terminal": False,
-        }
+        state = EnvState(
+            oxygen=params.max_oxygen,
+            sub_x=5,
+            sub_y=0,
+            sub_or=0,
+            f_bullet_count=0,
+            f_bullets=jnp.zeros((100, 3), dtype=jnp.int32),
+            e_bullet_count=0,
+            e_bullets=jnp.zeros((100, 3), dtype=jnp.int32),
+            e_fish_count=0,
+            e_fish=jnp.zeros((100, 5), dtype=jnp.int32),
+            e_subs_count=0,
+            e_subs=jnp.zeros((100, 5), dtype=jnp.int32),
+            diver_count=0,
+            divers=jnp.zeros((100, 4), dtype=jnp.int32),
+            e_spawn_speed=params.init_spawn_speed,
+            e_spawn_timer=params.init_spawn_speed,
+            d_spawn_timer=params.diver_spawn_speed,
+            move_speed=params.init_move_interval,
+            ramp_index=0,
+            shot_timer=0,
+            surface=1,
+            time=0,
+            terminal=False,
+        )
         return self.get_obs(state, params), state
 
-    def get_obs(self, state: dict, params: dict) -> Array:
+    def get_obs(self, state: EnvState, params: EnvParams) -> chex.Array:
         """Return observation from raw state trafo."""
         fish, sub, diver = [], [], []
         obs = jnp.zeros(self.obs_shape, dtype=bool)
         # Set agents sub-front and back, oxygen_gauge and diver_gauge
-        obs = jax.ops.index_update(
-            obs, jax.ops.index[state["sub_y"], state["sub_x"], 0], 1
-        )
-        back_x = (state["sub_x"] - 1) * state["sub_or"] + (
-            state["sub_x"] + 1
-        ) * (1 - state["sub_or"])
-        obs = jax.ops.index_update(
-            obs, jax.ops.index[state["sub_y"], back_x, 1], 1
-        )
-        obs = jax.ops.index_update(
-            obs,
-            jax.ops.index[
-                9, 0 : state["oxygen"] * 10 // params["max_oxygen"], 7
-            ],
-            1,
-        )
-        obs = jax.ops.index_update(
-            obs, jax.ops.index[9, 9 - state["diver_count"] : 9, 8], 1
+        obs = obs.at[state.sub_y, state.sub_x, 0].set(1)
+        back_x = (state.sub_x - 1) * state.sub_or + (state.sub_x + 1) * (
+            1 - state.sub_or
         )
+        obs = obs.at[state.sub_y, back_x, 1].set(1)
+        obs = obs.at[9, 0 : state.oxygen * 10 // params.max_oxygen, 7].set(1)
+        obs = obs.at[9, 9 - state.diver_count : 9, 8].set(1)
 
         # Set friendly bulltes, enemy bullets, enemy fish+trail, enemey sub+trail
-        for f_b_id in range(state["f_bullet_count"]):
-            obs = jax.ops.index_update(
-                obs,
-                jax.ops.index[
-                    state["f_bullets"][f_b_id, 1],
-                    state["f_bullets"][f_b_id, 0],
-                    2,
-                ],
-                1,
-            )
-        for e_b_id in range(state["e_bullet_count"]):
-            obs = jax.ops.index_update(
-                obs,
-                jax.ops.index[
-                    state["e_bullets"][e_b_id, 1],
-                    state["e_bullets"][e_b_id, 0],
-                    4,
-                ],
-                1,
-            )
-
-        for e_f_id in range(state["e_fish_count"]):
-            obs = jax.ops.index_update(
-                obs,
-                jax.ops.index[
-                    state["e_fish"][e_f_id, 1], state["e_fish"][e_f_id, 0], 5
-                ],
-                1,
-            )
+        for f_b_id in range(state.f_bullet_count):
+            obs = obs.at[
+                state.f_bullets[f_b_id, 1],
+                state.f_bullets[f_b_id, 0],
+                2,
+            ].set(1)
+        for e_b_id in range(state.e_bullet_count):
+            obs = obs.at[
+                state.e_bullets[e_b_id, 1],
+                state.e_bullets[e_b_id, 0],
+                4,
+            ].set(1)
+        for e_f_id in range(state.e_fish_count):
+            obs = obs.at[
+                state.e_fish[e_f_id, 1], state.e_fish[e_f_id, 0], 5
+            ].set(1)
             back_x = (fish[0] - 1) * fish[2] + (fish[0] + 1) * (1 - fish[2])
             border_cond = jnp.logical_and(back_x >= 0, back_x <= 9)
-            obs = (
-                border_cond
-                * jax.ops.index_update(
-                    obs, jax.ops.index[state["e_fish"][e_f_id][1], back_x, 3], 1
-                )
-                + (1 - border_cond) * obs
+            obs = jax.lax.select(
+                border_cond,
+                obs.at[state.e_fish[e_f_id][1], back_x, 3].set(1),
+                obs,
             )
 
-        for e_s_id in range(state["e_subs_count"]):
-            obs = jax.ops.index_update(
-                obs,
-                jax.ops.index[
-                    state["e_subs"][e_s_id, 1], state["e_subs"][e_s_id, 0], 6
-                ],
-                1,
-            )
+        for e_s_id in range(state.e_subs_count):
+            obs = obs.at[
+                state.e_subs[e_s_id, 1], state.e_subs[e_s_id, 0], 6
+            ].set(1)
             back_x = (sub[0] - 1) * sub[2] + (sub[0] + 1) * (1 - sub[2])
             border_cond = jnp.logical_and(back_x >= 0, back_x <= 9)
-            obs = (
-                border_cond
-                * jax.ops.index_update(
-                    obs, jax.ops.index[state["e_subs"][e_s_id, 1], back_x, 3], 1
-                )
-                + (1 - border_cond) * obs
+            obs = jax.lax.select(
+                border_cond,
+                obs.at[state.e_subs[e_s_id, 1], back_x, 3].set(1),
+                obs,
             )
 
-        for d_id in range(state["diver_count"]):
-            obs = jax.ops.index_update(
-                obs,
-                jax.ops.index[
-                    state["divers"][d_id, 1], state["divers"][d_id, 0], 9
-                ],
-                1,
-            )
+        for d_id in range(state.diver_count):
+            obs = obs.at[state.divers[d_id, 1], state.divers[d_id, 0], 9].set(1)
             back_x = (diver[0] - 1) * diver[2] + (diver[0] + 1) * (1 - diver[2])
             border_cond = jnp.logical_and(back_x >= 0, back_x <= 9)
-            obs = (
-                border_cond
-                * jax.ops.index_update(
-                    obs, jax.ops.index[state["divers"][d_id, 1], back_x, 3], 1
-                )
-                + (1 - border_cond) * obs
+            obs = jax.lax.select(
+                border_cond,
+                obs.at[state.divers[d_id, 1], back_x, 3].set(1),
+                obs,
             )
-        return obs
+        return obs.astype(jnp.float32)
 
-    def is_terminal(self, state: dict, params: dict) -> bool:
+    def is_terminal(self, state: EnvState, params: EnvParams) -> bool:
         """Check whether state is terminal."""
-        done_steps = state["time"] > params["max_steps_in_episode"]
-        return jnp.logical_or(state["terminal"], done_steps)
+        done_steps = state.time >= params.max_steps_in_episode
+        return jnp.logical_or(state.terminal, done_steps)
 
     @property
     def name(self) -> str:
@@ -259,19 +251,21 @@ class MinSeaquest(environment.Environment):
         """Number of actions possible in environment."""
         return len(self.action_set)
 
-    def action_space(self, params) -> spaces.Discrete:
+    def action_space(
+        self, params: Optional[EnvParams] = None
+    ) -> spaces.Discrete:
         """Action space of the environment."""
         return spaces.Discrete(len(self.action_set))
 
-    def observation_space(self, params: dict):
+    def observation_space(self, params: EnvParams) -> spaces.Box:
         """Observation space of the environment."""
         return spaces.Box(0, 1, self.obs_shape)
 
-    def state_space(self, params: dict):
+    def state_space(self, params: EnvParams) -> spaces.Dict:
         """State space of the environment."""
         return spaces.Dict(
             {
-                "oxygen": spaces.Discrete(params["max_oxygen"]),
+                "oxygen": spaces.Discrete(params.max_oxygen),
                 "diver_count": spaces.Discrete(20),
                 "sub_x": spaces.Discrete(10),
                 "sub_y": spaces.Discrete(10),
@@ -281,20 +275,20 @@ class MinSeaquest(environment.Environment):
                 "e_fish": spaces.Box(0, 1, (100, 5)),
                 "e_subs": spaces.Box(0, 1, (100, 5)),
                 "divers": spaces.Box(0, 1, (100, 4)),
-                "e_spawn_speed": spaces.Discrete(params["init_spawn_speed"]),
-                "e_spawn_timer": spaces.Discrete(params["init_spawn_speed"]),
-                "d_spawn_timer": spaces.Discrete(params["diver_spawn_speed"]),
+                "e_spawn_speed": spaces.Discrete(params.init_spawn_speed),
+                "e_spawn_timer": spaces.Discrete(params.init_spawn_speed),
+                "d_spawn_timer": spaces.Discrete(params.diver_spawn_speed),
                 "move_speed": spaces.Discrete(1000),
                 "ramp_index": spaces.Discrete(1000),
-                "shot_timer": spaces.Discrete(params["shot_cool_down"]),
+                "shot_timer": spaces.Discrete(params.shot_cool_down),
                 "surface": spaces.Discrete(2),
-                "time": spaces.Discrete(params["max_steps_in_episode"]),
+                "time": spaces.Discrete(params.max_steps_in_episode),
                 "terminal": spaces.Discrete(2),
             }
         )
 
 
-def step_agent(state: dict, action: int, env_params: dict) -> dict:
+def step_agent(state: EnvState, action: int, env_params: EnvParams) -> EnvState:
     """Perform submarine position and friendly bullets transition."""
     # Update submarine position based on l, r or u, d actions
     not_l_or_r = jnp.logical_and(action != 1, action != 3)
@@ -330,9 +324,9 @@ def step_agent(state: dict, action: int, env_params: dict) -> dict:
     return state
 
 
-def step_bullets(state: dict) -> Tuple[dict, float]:
+def step_bullets(state: EnvState) -> Tuple[EnvState, float]:
     """Perform friendly bullets transition."""
-    reward = 0
+    reward = 0.0
     f_bullet_count = 0
     f_bullets = jnp.zeros((100, 3))
     e_fish = jnp.zeros((100, 3))
@@ -395,7 +389,9 @@ def step_e_bullets():
     return
 
 
-def spawn_enemy(key: PRNGKey, state: dict, env_params: dict) -> Array:
+def spawn_enemy(
+    key: chex.PRNGKey, state: EnvState, env_params: EnvParams
+) -> chex.Array:
     """Spawn a new enemy."""
     lr_key, sub_key, y_key = jax.random.splt(key, 3)
     lr = jax.random.choice(lr_key, 2, ())
@@ -422,76 +418,69 @@ def spawn_enemy(key: PRNGKey, state: dict, env_params: dict) -> Array:
     )
 
 
-def spawn_diver(key: PRNGKey, env_params: dict) -> Array:
+def spawn_diver(key: chex.PRNGKey, env_params: EnvParams) -> chex.Array:
     """Spawn a new diver."""
     lr_key, y_key = jax.random.splt(key)
     lr = jax.random.choice(lr_key, 2, ())
     x = lax.select(lr, 0, 9)
     y = jax.random.choice(y_key, jnp.arange(1, 9), ())
-    return jnp.array([x, y, lr, env_params["diver_move_interval"]])
+    return jnp.array([x, y, lr, env_params.diver_move_interval])
 
 
-def step_timers(state, reward, env_params):
+def step_timers(state: EnvState, reward: float, env_params: EnvParams):
     """Update the timers of the environment and calculate surface reward."""
-    state["e_spawn_timer"] -= state["e_spawn_timer"] > 0
-    state["d_spawn_timer"] -= state["d_spawn_timer"] > 0
-    state["shot_timer"] -= state["shot_timer"] > 0
-    oxy_term = lax.select(state["oxygen"] < 0, 1, 0)
+    e_spawn_timer = state.e_spawn_timer - state.e_spawn_timer > 0
+    d_spawn_timer = state.d_spawn_timer - state.d_spawn_timer > 0
+    shot_timer = state.shot_timer - state.shot_timer > 0
+    oxy_term = lax.select(state.oxygen < 0, 1, 0)
 
     # Update oxygen and surface indicator if submarine is above
-    above_surface = state["sub_y"] > 0
-    state["oxygen"] = lax.select(
-        above_surface, state["oxygen"] - 1, state["oxygen"]
-    )
-    state["surface"] = lax.select(above_surface, 1, 0)
+    above_surface = state.sub_y > 0
+    oxygen = lax.select(above_surface, state.oxygen - 1, state.oxygen)
+    surface = lax.select(above_surface, 1, 0)
 
     # Calculate reward/terminate episode otherwise
-    below_cond = jnp.logical_and(1 - above_surface, 1 - state["surface"])
-    diver_term = jnp.logical_and(below_cond, state["diver_count"] == 0)
-    surface_cond = jnp.logical_and(below_cond, 1 - (state["diver_count"] == 0))
+    below_cond = jnp.logical_and(1 - above_surface, 1 - surface)
+    diver_term = jnp.logical_and(below_cond, state.diver_count == 0)
+    surface_cond = jnp.logical_and(below_cond, 1 - (state.diver_count == 0))
     state, surface_reward = surface(surface_cond, state, env_params)
     reward += surface_cond * surface_reward
+    # TODO: Replace state ingredients
     return state, reward
 
 
 def surface(
-    surface_cond: bool, state: dict, env_params: dict
-) -> Tuple[dict, float]:
+    surface_cond: bool, state: EnvState, env_params: EnvParams
+) -> Tuple[EnvState, float]:
     """Perform surface transition and reward calculations."""
     surface = 1
-    diver_count = lax.select(state["diver_count"] == 6, 0, state["diver_count"])
+    diver_count = lax.select(state.diver_count == 6, 0, state.diver_count)
     reward = lax.select(
-        state["diver_count"] == 6,
-        state["oxygen"] * 10 // env_params["max_oxygen"],
+        state.diver_count == 6,
+        state.oxygen * 10 // env_params.max_oxygen,
         0,
     )
-    oxygen = env_params["oxygen"]
+    oxygen = state.oxygen
     diver_count -= 1
     ramp_cond = jnp.logical_and(
-        env_params["ramping"],
-        jnp.logical_or(state["e_spawn_speed"] > 1, state["move_speed"] > 2),
+        env_params.ramping,
+        jnp.logical_or(state.e_spawn_speed > 1, state.move_speed > 2),
     )
     move_cond = jnp.logical_and(
         ramp_cond,
-        jnp.logical_and(state["move_speed"] > 2, state["ramp_index"] % 2),
+        jnp.logical_and(state.move_speed > 2, state.ramp_index % 2),
     )
-    move_speed = lax.select(
-        move_cond, state["move_speed"] - 1, state["move_speed"]
-    )
-    e_spawn_cond = jnp.logical_and(ramp_cond, state["e_spawn_speed"] > 1)
+    move_speed = lax.select(move_cond, state.move_speed - 1, state.move_speed)
+    e_spawn_cond = jnp.logical_and(ramp_cond, state.e_spawn_speed > 1)
     e_spawn_speed = lax.select(
-        e_spawn_cond, state["e_spawn_speed"] - 1, state["e_spawn_speed"]
+        e_spawn_cond, state.e_spawn_speed - 1, state.e_spawn_speed
     )
 
     # Update the state based on the surface_cond - only update if cond met!
-    state["diver_count"] = lax.select(
-        surface_cond, diver_count, state["diver_count"]
-    )
-    state["oxygen"] = lax.select(surface_cond, oxygen, state["oxygen"])
-    state["move_speed"] = lax.select(
-        surface_cond, move_speed, state["move_speed"]
-    )
-    state["e_spawn_speed"] = lax.select(
-        surface_cond, e_spawn_speed, state["e_spawn_speed"]
+    state.diver_count = lax.select(surface_cond, diver_count, state.diver_count)
+    state.oxygen = lax.select(surface_cond, oxygen, state.oxygen)
+    state.move_speed = lax.select(surface_cond, move_speed, state.move_speed)
+    state.e_spawn_speed = lax.select(
+        surface_cond, e_spawn_speed, state.e_spawn_speed
     )
     return state, reward
