@@ -1,5 +1,5 @@
-from gymnax.utils.state_translate import np_state_to_jax
-from gymnax.utils.test_helpers import (
+from .state_translate import np_state_to_jax
+from .test_helpers import (
     assert_correct_state,
     assert_correct_transit,
     minatar_action_map,
